# -*- coding: UTF-8 -*-
from .dataset import IcvDataSet
import os
import shutil
import numpy as np
from ..utils import load_voc_anno, save_voc_anno, make_empty_voc_anno, fcopy, list_from_file, list_to_file, is_file, \
    is_dir, mkdir
from ..image import imread
from ..data.core.meta import SampleMeta, AnnoMeta
from ..data.core.sample import Sample, Anno
from ..data.core.bbox import BBox
from ..data.core.mask import Mask
from ..data.core.polys import Polygon
from tqdm import tqdm
import random
from ..vis.color import STANDARD_COLORS


# TODO: 分割数据（图像目录、segment=1）
class Voc(IcvDataSet):
    def __init__(self, data_dir, split=None, use_difficult=True, keep_no_anno_image=True, mode="detect",
                 include_segment=True, categories=None, one_index=False, image_suffix="jpg"):
        self.root = data_dir
        self.split = split if split else "trainval"
        self.mode = mode
        self.is_seg_mode = self.mode != "detect"
        self.include_segment = include_segment
        self.image_set = "Main" if self.mode == "detect" else "Segmentation"

        self.use_difficult = use_difficult
        self.keep_no_anno_image = keep_no_anno_image
        self.image_suffix = image_suffix
        self.one_index = one_index

        self._annopath = os.path.join(self.root, "Annotations", "%s.xml")
        self._imgpath = os.path.join(self.root, "JPEGImages", "%s." + image_suffix)
        self._seg_class_imgpath = os.path.join(self.root, "SegmentationClass", "%s.png")
        self._seg_object_imgpath = os.path.join(self.root, "SegmentationObject", "%s.png")
        self._imgsetpath = os.path.join(self.root, "ImageSets", self.image_set, "%s.txt")

        self.categories = categories

        self.init()

        super(Voc, self).__init__(self.ids, self.categories, self.keep_no_anno_image, one_index)

    def init(self):
        self.ids = list_from_file(self._imgsetpath % self.split)
        self.id2img = {k: v for k, v in enumerate(self.ids)}

        self.sample_db = {}
        self.color_map = {}
        if self.categories is None:
            print("parsing categories ...")
            self.categories = self.get_categories()

        self.set_categories(self.categories)
        self.set_colormap(self.color_map)
        print("there have %d samples in VOC dataset" % len(self.ids))
        print("there have %d categories in VOC dataset" % len(self.categories))

    def concat(self, voc, output_dir, reset=False, new_split=None):
        assert isinstance(voc, Voc)
        if new_split is None:
            if self.split == voc.split:
                new_split = self.split
            else:
                new_split = self.split + voc.split

        anno_path, image_path, imgset_path, imgset_seg_path, seg_class_image_path, seg_object_image_path = Voc.reset_dir(
            output_dir, reset=reset)

        ps0 = fcopy([self._imgpath % id for id in self.ids], image_path)
        ps1 = fcopy([self._imgpath % id for id in voc.ids], image_path)

        ids0 = [os.path.basename(_).rsplit(".", 1)[0] for _ in ps0]
        ids1 = [os.path.basename(_).rsplit(".", 1)[0] for _ in ps1]

        fcopy([self._annopath % id for id in self.ids], anno_path)
        fcopy([voc._annopath % id for id in voc.ids], anno_path)

        if self.is_seg_mode:
            fcopy([self._seg_class_imgpath % id for id in self.ids], seg_class_image_path)
            fcopy([self._seg_object_imgpath % id for id in self.ids], seg_object_image_path)

        if voc.is_seg_mode:
            fcopy([voc._seg_class_imgpath % id for id in voc.ids], seg_class_image_path)
            fcopy([voc._seg_object_imgpath % id for id in voc.ids], seg_object_image_path)

        if self.mode == voc.mode:
            ids = ids0 + ids1
            setpath = imgset_seg_path if self.is_seg_mode else imgset_path

            list_to_file(ids, os.path.join(setpath, "%s.txt" % new_split))
        else:
            if not self.is_seg_mode:
                list_to_file(ids0, os.path.join(imgset_path, "%s.txt" % new_split))
            else:
                list_to_file(ids0, os.path.join(imgset_seg_path, "%s.txt" % new_split))

            if not voc.is_seg_mode:
                list_to_file(ids1, os.path.join(imgset_path, "%s.txt" % new_split))
            else:
                list_to_file(ids1, os.path.join(imgset_seg_path, "%s.txt" % new_split))

        return Voc(
            output_dir,
            split=new_split,
            use_difficult=self.use_difficult,
            keep_no_anno_image=self.keep_no_anno_image,
            mode=self.mode,
            include_segment=self.include_segment,
            one_index=self.one_index,
            categories=self.categories + voc.categories,
            image_suffix=self.image_suffix
        )

    def sub(self, count=0, ratio=0, shuffle=True, output_dir=None):
        voc = super(Voc, self).sub(count, ratio, shuffle, output_dir)
        voc.root = output_dir
        return voc

    def save(self, output_dir, reset_dir=False, split=None):
        split = split if split is not None else self.split
        anno_path, image_path, imgset_path, imgset_seg_path, seg_class_image_path, seg_object_image_path = Voc.reset_dir(
            output_dir, reset=reset_dir)
        for id in self.ids:
            if is_file(self._annopath % id) and is_file(self._imgpath % id):
                fcopy(self._annopath % id, anno_path)
                fcopy(self._imgpath % id, image_path)
                fcopy(self._seg_class_imgpath % id, seg_class_image_path)
                fcopy(self._seg_object_imgpath % id, seg_object_image_path)
            else:
                self._write_sample(self.get_sample(id), output_dir)

        if self.is_seg_mode:
            list_to_file(self.ids, os.path.join(imgset_seg_path, "%s.txt" % split))
        else:
            list_to_file(self.ids, os.path.join(imgset_path, "%s.txt" % split))

    @staticmethod
    def reset_dir(dist_dir, reset=False):
        if not reset:
            assert is_dir(dist_dir)
        if reset and os.path.exists(dist_dir):
            shutil.rmtree(dist_dir)

        anno_path = os.path.join(dist_dir, "Annotations")
        image_path = os.path.join(dist_dir, "JPEGImages")
        imgset_path = os.path.join(dist_dir, "ImageSets", "Main")

        imgset_seg_path = os.path.join(dist_dir, "ImageSets", "Segmentation")
        seg_class_image_path = os.path.join(dist_dir, "SegmentationClass")
        seg_object_image_path = os.path.join(dist_dir, "SegmentationObject")

        for _path in [anno_path, image_path, imgset_path, imgset_seg_path, seg_class_image_path, seg_object_image_path]:
            if reset or not is_dir(_path):
                mkdir(_path)

        return anno_path, image_path, imgset_path, imgset_seg_path, seg_class_image_path, seg_object_image_path

    def get_categories(self):
        self.get_samples()
        categories = []
        for anno_sample in self.samples:
            label_list = [l.label for l in anno_sample.annos if l.label is not None]
            if label_list:
                categories.extend(label_list)
        categories = list(set(categories))
        categories.sort()
        return categories

    def get_sample(self, id):
        if id in self.sample_db:
            return self.sample_db[id]

        anno_file = self._annopath % id
        image_file = self._imgpath % id

        if not is_file(image_file):
            raise FileNotFoundError("image file : {} not exist!".format(image_file))

        if not is_file(anno_file):
            anno_data = make_empty_voc_anno()
        else:
            anno_data = load_voc_anno(anno_file)["annotation"]

        sample_meta = SampleMeta({
            k: anno_data[k]
            for k in anno_data if k not in ["object"]
        })

        segcls_file = self._seg_class_imgpath % id
        segobj_file = self._seg_object_imgpath % id

        img_segcls = None
        if is_file(segcls_file):
            img_segcls = imread(segcls_file, 0)

        img_segobj = None
        if is_file(segobj_file):
            img_segobj = imread(segobj_file, 0)

        annos = []
        if "object" in anno_data:
            for obj in anno_data["object"]:
                if "difficult" in obj and obj["difficult"] != '0' and self.use_difficult:
                    continue
                label = obj["name"]
                if label not in self.color_map:
                    self.color_map[label] = random.choice(STANDARD_COLORS)

                xmin = int(obj["bndbox"]["xmin"])
                ymin = int(obj["bndbox"]["ymin"])
                xmax = int(obj["bndbox"]["xmax"])
                ymax = int(obj["bndbox"]["ymax"])
                anno = Anno(
                    bbox=BBox(
                        xmin=xmin,
                        ymin=ymin,
                        xmax=xmax,
                        ymax=ymax,
                        label=label,
                    ),
                    label=label,
                    color=self.color_map[label],
                    meta=AnnoMeta({
                        k: obj[k]
                        for k in obj if k not in ["name", "bndbox"]
                    })
                )

                if img_segobj is not None:
                    bin_mask = np.zeros_like(img_segobj, np.uint8)
                    bin_mask[ymin:ymax, xmin:xmax] = img_segobj[ymin:ymax, xmin:xmax]
                    bin_mask[np.where(bin_mask != 0)] = 1
                    anno.mask = Mask(bin_mask)
                elif img_segcls is not None:
                    bin_mask = np.zeros_like(img_segcls, np.uint8)
                    bin_mask[ymin:ymax, xmin:xmax] = img_segcls[ymin:ymax, xmin:xmax]
                    bin_mask[np.where(bin_mask != 0)] = 1
                    anno.mask = Mask(bin_mask)

                annos.append(anno)

        sample = Sample(
            id,
            image_file,
            annos,
            sample_meta
        )

        self.sample_db[id] = sample
        return sample

    def _write_sample(self, sample, dist_dir):
        assert isinstance(sample, Sample)

        reset_dir = not os.path.exists(dist_dir)
        anno_path, image_path, imgset_path, imgset_seg_path, \
        seg_class_image_path, seg_object_image_path = Voc.reset_dir(dist_dir, reset_dir)

        anno_dist_path = os.path.join(anno_path, "%s.xml" % sample.name)
        anno_data = {
            "segmented": 1 if self.is_seg_mode else 0,
            "folder": image_path,
            "filename": os.path.basename(sample.path),
            "size": {
                "width": sample.width,
                "height": sample.height,
                "depth": sample.dim
            },
            "objects": []
        }

        for ix, anno in enumerate(sample.annos):
            bndbox = anno.meta.dict()
            if "truncated" not in bndbox and "difficult" not in bndbox:
                bndbox = {
                    "truncated": 1,
                    "difficult": 0,
                }
            bndbox["name"] = anno.label
            bndbox["bndbox"] = {
                "xmin": anno.bbox.xmin_int,
                "ymin": anno.bbox.ymin_int,
                "xmax": anno.bbox.xmax_int,
                "ymax": anno.bbox.ymax_int,
            }

            catid = self.cat2id[anno.label]
            if not self.one_index:
                catid += 1

            if anno.seg_mode_mask:
                anno.mask.save(os.path.join(seg_class_image_path, "%s.png" % sample.name), id=catid)
                # TODO: make sure object segment mask id
                anno.mask.save(os.path.join(seg_object_image_path, "%s.png" % sample.name), id=ix + 1)
            elif anno.seg_mode_polys:
                assert isinstance(anno.polys, Polygon)
                anno.polys.to_mask(sample.height, sample.width).save(
                    os.path.join(seg_class_image_path, "%s.png" % sample.name),
                    id=catid)
                # TODO: make sure object segment mask id
                anno.polys.to_mask(sample.height, sample.width).save(
                    os.path.join(seg_object_image_path, "%s.png" % sample.name), id=ix + 1)

            anno_data["objects"].append(bndbox)

        save_voc_anno(anno_data, anno_dist_path)
        fcopy(sample.path, image_path)

    def vis(self, id=None, with_bbox=True, with_seg=True, is_show=False, save_dir=None, reset_dir=False):
        if save_dir is not None:
            if not os.path.exists(save_dir):
                mkdir(save_dir)
            elif reset_dir:
                shutil.rmtree(save_dir)
                mkdir(save_dir)

        if id is not None:
            sample = self.get_sample(id)
            save_path = None if save_dir is None else os.path.join(save_dir, "%s.jpg" % sample.name)
            return sample.vis(with_bbox=with_bbox, with_seg=with_seg, is_show=is_show, save_path=save_path)

        image_vis = []
        for id in tqdm(self.ids):
            sample = self.get_sample(id)
            save_path = None if save_dir is None else os.path.join(save_dir, "%s.jpg" % sample.name)
            image = sample.vis(with_bbox=with_bbox, with_seg=with_seg, is_show=is_show, save_path=save_path)
            image_vis.append(image)
        return image_vis



# -*- coding: UTF-8 -*-
from icv.utils import is_valid_url
from urllib import request
from urllib.error import URLError,HTTPError
from urllib.parse import urlencode
from .response import IcvHttpResponse

class IcvHttpRequest(request.Request):
    method = "GET"
    url = ""
    data = {}
    params = {}
    headers = {}
    timeout = 3000

    def __init__(self,method,url,data=None,params=None,headers=None,timeout=3000):
        assert is_valid_url(url)
        self.method = method
        self.url = url
        self.data = data
        self.params = params if params else {}
        self.headers = headers if headers else {}
        self.timeout = timeout

        if self.params is not None:
            params = urlencode(self.params)
            self.url = "?".join([self.url,params])

        super(IcvHttpRequest,self).__init__(self.url,data=self.data,headers=self.headers,method=self.method)

    def __repr__(self):
        return "url:{}\nmethod:{}\ndata:{}\nparams:{}\nheaders:{}\ntimeout:{}".format(
            self.url,self.method,self.data,self.params,self.headers,self.timeout
        )

    def send(self):
        req = request.Request(self.url,data=self.data,headers=self.headers,method=self.method)
        response = IcvHttpResponse(content="",status_code=0)
        try:
            res = request.urlopen(req,timeout=self.timeout)
            response.content = res.read().decode("utf-8")
            response.status_code = 200
            response.headers = res.info()
        except HTTPError as e:
            if hasattr(e,"headers"):
                response.headers = e.headers
            if hasattr(e, 'reason'):
                response.error = e.reason

            response.content = str(e)
            response.status_code = 500
            if hasattr(e, 'code'):
                response.status_code = e.code

        except URLError as e:
            if hasattr(e,"headers"):
                response.headers = e.headers
            if hasattr(e, 'reason'):
                response.error = e.reason

            response.content = str(e)
            response.status_code = 404
            if hasattr(e, 'code'):
                response.status_code = e.code

        return response







# -*- coding: UTF-8 -*-
from ..utils import is_dir, list_from_file
from ..image.io import imwrite
from .core import Image
import os
import shutil

class Classify(object):
    def __init__(self, root_dir=None, sample_file_list=None, sample_split="\t", categories=None, one_index=False):
        assert is_dir(root_dir) or sample_file_list is not None

        if is_dir(root_dir):
            self.root_dir = root_dir
            self.img2cat, self.img2set = Classify.extract_by_dir(self.root_dir)
        else:
            self.img2cat = {}
            self.img2set = {}
            self.categories = []

            for sample_file in sample_file_list:
                img2cat, img2set = Classify.extract_by_samplefile(sample_file, sample_split)

                self.img2cat.update(img2cat)
                self.img2set.update(img2set)

        if categories is not None:
            self.categories = categories
        else:
            self.categories = list(set(self.img2cat.values()))

    def save(self, output, rename=False, suffix=None):
        catidx = {}
        for img in self.img2set:
            set_dir = os.path.join(output, self.img2set[img])
            cat = self.img2cat[img]
            if cat not in catidx:
                catidx[cat] = -1

            catidx[cat] += 1

            if not is_dir(set_dir):
                os.makedirs(set_dir)

            cat_dir = os.path.join(set_dir, cat)
            if not is_dir(cat_dir):
                os.makedirs(cat_dir)

            if isinstance(img,Image):
                name, ext = img.name,img.ext
            else:
                name, ext = os.path.basename(img).rsplit(".", 1)
            suffix = ext if suffix is None else suffix.strip(".")

            new_name = "{}_{}.{}".format(cat, catidx[cat], suffix) if rename else "{}.{}".format(name, suffix)

            if isinstance(img,Image):
                imwrite(img.img,os.path.join(cat_dir, new_name))
            else:
                shutil.copy(img, os.path.join(cat_dir, new_name))

    @staticmethod
    def extract_by_dir(root_dir):
        img2cat = {}
        img2set = {}
        for s in os.listdir(root_dir):
            set_dir = os.path.join(root_dir, s)
            if is_dir(set_dir):
                for cat in os.listdir(set_dir):
                    cat_dir = os.path.join(set_dir, cat)
                    if is_dir(cat_dir):
                        for f in cat_dir:
                            img2cat[os.path.join(cat_dir, f)] = cat
                            img2set[os.path.join(cat_dir, f)] = s

        return img2cat, img2set

    @staticmethod
    def extract_by_samplefile(samplefile, split="\t"):
        samples = list_from_file(samplefile)
        _set = os.path.basename(samplefile).rsplit(".", 1)[0]

        img2cat = {}
        img2set = {}

        for s in samples:
            ss = s.split(split)
            img2cat[ss[0]] = ss[1]

            img2set[ss[0]] = _set

        return img2cat, img2set



# -*- coding: utf-8 -* -
'''
Pascal Voc 分割图片相关操作
'''
import cv2
import numpy as np
from PIL import Image, ImageDraw

def getsegmentation(seg_img_path, bbox):
    '''
    获取分割图片区域的mask（pascal voc数据中，每个分割区域都有一个检测框与之对应）
    :param seg_img_path: 分割的png图片地址
    :param bbox: 分割区域对应的框的坐标[xmin ymin xmax ymax]
    :return:
    '''
    try:
        mask_1 = cv2.imread(seg_img_path, 0)
        mask = np.zeros_like(mask_1, np.uint8)
        rectangle = bbox
        mask[rectangle[1]:rectangle[3], rectangle[0]:rectangle[2]] = mask_1[rectangle[1]:rectangle[3],
                                                                     rectangle[0]:rectangle[2]]

        # 计算矩形中点像素值
        mean_x = (rectangle[0] + rectangle[2]) // 2
        mean_y = (rectangle[1] + rectangle[3]) // 2

        end = min((mask.shape[1], int(rectangle[2]) + 1))
        start = max((0, int(rectangle[0]) - 1))

        flag = True
        for i in range(mean_x, end):
            x_ = i
            y_ = mean_y
            pixels = mask_1[y_, x_]
            if pixels != 0 and pixels != 220:  # 0 对应背景 220对应边界线
                mask = (mask == pixels).astype(np.uint8)
                flag = False
                break
        if flag:
            for i in range(mean_x, start, -1):
                x_ = i
                y_ = mean_y
                pixels = mask_1[y_, x_]
                if pixels != 0 and pixels != 220:
                    mask = (mask == pixels).astype(np.uint8)
                    break
        return mask

    except Exception as e:
        print("getsegmentation error:",e)
        return [0]

def getsegmask(seg_img_mask, bbox):
    '''
    获取分割图片区域的mask列表（pascal voc数据中，每个分割区域都有一个检测框与之对应）
    :param seg_img_mask: 分割的png图片对应的numpy数组，使用cv2.read("/path/to/image",0)
    :param bbox: 分割区域对应的框的坐标列表[[xmin1 ymin1 xmax1 ymax1],[xmin2 ymin2 xmax2 ymax2],...]
    :return:
    '''
    try:
        mask = np.zeros_like(seg_img_mask, np.uint8)
        rectangle = bbox
        mask[rectangle[1]:rectangle[3], rectangle[0]:rectangle[2]] = seg_img_mask[rectangle[1]:rectangle[3],
                                                                     rectangle[0]:rectangle[2]]

        # 计算矩形中点像素值
        mean_x = (rectangle[0] + rectangle[2]) // 2
        mean_y = (rectangle[1] + rectangle[3]) // 2

        end = min((mask.shape[1], int(rectangle[2]) + 1))
        start = max((0, int(rectangle[0]) - 1))

        flag = True
        for i in range(mean_x, end):
            x_ = i
            y_ = mean_y
            pixels = seg_img_mask[y_, x_]
            if pixels != 0 and pixels != 220:  # 0 对应背景 220对应边界线
                mask = (mask == pixels).astype(np.uint8)
                flag = False
                break
        if flag:
            for i in range(mean_x, start, -1):
                x_ = i
                y_ = mean_y
                pixels = seg_img_mask[y_, x_]
                if pixels != 0 and pixels != 220:
                    mask = (mask == pixels).astype(np.uint8)
                    break
        return mask

    except Exception as e:
        print("getsegmentation error:",e)
        return [0]

def mask2box(mask):
    '''
    从mask反算出其边框
    mask：[h,w]  0、1组成的图片
    1对应对象，只需计算1对应的行列号（左上角行列号，右下角行列号，就可以算出其边框）
    :param mask:
    :return:
    '''
    # np.where(mask==1)
    index = np.argwhere(mask == 1)
    rows = index[:, 0]
    clos = index[:, 1]
    # 解析左上角行列号
    left_top_r = np.min(rows)  # y
    left_top_c = np.min(clos)  # x

    # 解析右下角行列号
    right_bottom_r = np.max(rows)
    right_bottom_c = np.max(clos)

    # [xmin,ymin,xmax,ymax]
    return [left_top_c, left_top_r, right_bottom_c, right_bottom_r]


def mask2polygons(mask):
    '''
    根据分割区域的mask计算对应的轮廓坐标点
    :param mask:
    :return:
    '''
    contours = cv2.findContours(mask, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)  # 找到轮廓线
    bbox = []
    for cont in contours[1]:
        cont_list = list(cont.flatten())
        for i in range(0,len(cont_list),2):
            # [x坐标,y坐标]
            bbox.append([cont_list[i],cont_list[i+1]])
        # [bbox.append(i) for i in list(cont.flatten())]
        # map(bbox.append,list(cont.flatten()))
    return bbox  # list(contours[1][0].flatten())

def polygons_to_mask(img_shape, polygons):
    '''
    根据图片的shape，以及每个分割区域的polygon获取mask
    :param img_shape:
    :param polygons:
    :return:
    '''
    mask = np.zeros(img_shape, dtype=np.uint8)
    mask = Image.fromarray(mask)
    # xy = list(map(tuple, polygons))
    xy = [tuple(p) for p in polygons]
    ImageDraw.Draw(mask).polygon(xy=xy, outline=1, fill=1)
    mask = np.array(mask, dtype=bool)
    return mask

def show_mask(img_path,polygons):
    '''
    显示分割区域
    :param img_path:
    :param polygons:
    :return:
    '''
    img = cv2.imread(img_path)
    height,width,_ = img.shape

    img = np.zeros([height,width],np.uint8)
    cv2.polylines(img, [np.asarray(polygons)], True, 1, lineType=cv2.LINE_AA)  # 画边界线
    cv2.fillPoly(img, [np.asarray(polygons)], 200)  # 画多边形 内部像素值为1

    cv2.imshow("",img)
    cv2.waitKey(0)
    cv2.destroyAllWindows()

def polygon_area(x, y):
    '''
    根据多边形的坐标列表计算面积
    :param x:
    :param y:
    :return:
    '''
    correction = x[-1] * y[0] - y[-1] * x[0]
    main_area = np.dot(x[:-1], y[1:]) - np.dot(y[:-1], x[1:])
    return 0.5 * np.abs(main_area + correction)





# -*- coding: UTF-8 -*-
import os
from ..image import imread, imwrite, imresize, imshow
from .detector import Detector
from ..data.core import BBox
from ..utils import Timer, is_file, is_seq
from .result import DetectionResult
try:
    import tensorflow as tf
    import numpy as np
except ImportError as e:
    print("you are not install tensorflow or numpy")
from .utils import ops as utils_ops

class TfObjectDetector(Detector):
    def __init__(self, model_path, labelmap_path=None, categories=[], iou_thr=0.5, score_thr=0.5, device=None):
        assert is_file(model_path)
        assert (is_seq(categories) and len(categories) > 0) or is_file(labelmap_path)
        super(TfObjectDetector, self).__init__(categories=categories, labelmap_path=labelmap_path, iou_thr=iou_thr,
                                               score_thr=score_thr, device=device)

        self.model_path = model_path
        self._build_detector()

    def _build_detector(self):
        self._load_model()
        self._load_sess()

    def _load_model(self):
        '''
        加载模型
        :return:
        '''
        detection_graph = tf.Graph()
        with detection_graph.as_default():
            od_graph_def = tf.GraphDef()
            with tf.gfile.GFile(self.model_path, 'rb') as fid:
                serialized_graph = fid.read()
                od_graph_def.ParseFromString(serialized_graph)
                tf.import_graph_def(od_graph_def, name='')
        self.graph = detection_graph

    def _load_sess(self):
        '''
        初始化session
        :return:
        '''
        with self.graph.as_default():
            self.sess = tf.Session()
            ops = tf.get_default_graph().get_operations()
            all_tensor_names = {output.name for op in ops for output in op.outputs}
            self.tensor_dict = {}
            for key in [
                'num_detections', 'detection_boxes', 'detection_scores',
                'detection_classes', 'detection_masks'
            ]:
                tensor_name = key + ':0'
                if tensor_name in all_tensor_names:
                    self.tensor_dict[key] = tf.get_default_graph().get_tensor_by_name(
                        tensor_name)

            self.image_tensor = tf.get_default_graph().get_tensor_by_name('image_tensor:0')

    def _reframe_detection_mask(self, image_np):
        '''
        对图片的分割掩码进行转换
        :param image:
        :return:
        '''
        if 'detection_masks' in self.tensor_dict:
            # The following processing is only for single image
            detection_boxes = tf.squeeze(self.tensor_dict['detection_boxes'], [0])
            detection_masks = tf.squeeze(self.tensor_dict['detection_masks'], [0])
            # Reframe is required to translate mask from box coordinates to image coordinates and fit the image size.
            real_num_detection = tf.cast(self.tensor_dict['num_detections'][0], tf.int32)
            detection_boxes = tf.slice(detection_boxes, [0, 0], [real_num_detection, -1])
            detection_masks = tf.slice(detection_masks, [0, 0, 0], [real_num_detection, -1, -1])
            detection_masks_reframed = utils_ops.reframe_box_masks_to_image_masks(
                detection_masks, detection_boxes, image_np.shape[0], image_np.shape[1])
            detection_masks_reframed = tf.cast(
                tf.greater(detection_masks_reframed, 0.5), tf.uint8)
            # Follow the convention by adding back the batch dimension
            self.tensor_dict['detection_masks'] = tf.expand_dims(
                detection_masks_reframed, 0)

    def inference(self, image, is_show=False, save_path=None, score_thr=-1):
        image_np = imread(image)
        img_height, img_width = image_np.shape[:2]
        self._reframe_detection_mask(image_np)
        timer = Timer()
        # Run inference
        output_dict = self.sess.run(self.tensor_dict,
                                    feed_dict={self.image_tensor: np.expand_dims(image_np, 0)})
        inference_time = timer.since_start()

        # all outputs are float32 numpy arrays, so convert types as appropriate
        num_detections = int(output_dict['num_detections'][0])
        detection_classes = output_dict[
            'detection_classes'][0].astype(np.uint8)
        # ymin,xmin,ymax,xmax
        detection_boxes = output_dict['detection_boxes'][0]
        detection_scores = output_dict['detection_scores'][0]
        detection_masks = None
        if 'detection_masks' in output_dict:
            detection_masks = output_dict['detection_masks'][0]

        detection_bboxes = [
            BBox(
                xmin=det_box[1] * img_width,
                ymin=det_box[0] * img_height,
                xmax=det_box[3] * img_width,
                ymax=det_box[2] * img_height
            )
            for det_box in detection_boxes
        ]

        score_thr = score_thr if score_thr >= 0 else self.score_thr
        det_result = DetectionResult(
            det_bboxes=detection_bboxes,
            det_classes=detection_classes,
            det_scores=detection_scores,
            det_masks=detection_masks,
            det_time=inference_time,
            categories=self.categories,
            score_thr=score_thr
        )

        det_result.vis(image_np)

        if is_show:
            imshow(det_result.det_image)

        if save_path is not None:
            imwrite(det_result.det_image, save_path)

        return det_result

    def inference_batch(self, image_batch, save_dir=None, resize=None, score_thr=-1):
        self._reframe_detection_mask(imread(image_batch[0]))
        if isinstance(image_batch, np.ndarray):
            image_np_batch = image_batch
        else:
            image_np_list = [imread(img) for img in image_batch]
            if len(set([img_np.shape for img_np in image_np_list])) != 1:
                if resize is None:
                    # inference one by one
                    return [self.inference(img_np) for img_np in image_np_list]
                    # raise Exception("if input image list not have same size, param resize must be a tuple of (width,height).")
            if resize is None:
                image_np_batch = np.array(image_np_list)
            else:
                image_np_batch = np.array(
                    [imresize(img_np, resize) for img_np in image_np_list]
                )

        # Run inference
        timer = Timer()
        output_dict = self.sess.run(self.tensor_dict,
                                    feed_dict={self.image_tensor: image_np_batch})

        inference_time = timer.since_start()

        score_thr = score_thr if score_thr >= 0 else self.score_thr
        det_result_list = []
        for ix, image in enumerate(image_np_batch):
            img_height, img_width = image.shape[:2]
            # all outputs are float32 numpy arrays, so convert types as appropriate
            num_detections = int(output_dict['num_detections'][ix])
            detection_classes = output_dict[
                'detection_classes'][ix].astype(np.uint8)
            detection_boxes = output_dict['detection_boxes'][ix]
            detection_scores = output_dict['detection_scores'][ix]
            detection_masks = None
            if 'detection_masks' in output_dict:
                detection_masks = output_dict['detection_masks'][ix]

            detection_bboxes = [
                BBox(
                    xmin=det_box[1] * img_width,
                    ymin=det_box[0] * img_height,
                    xmax=det_box[3] * img_width,
                    ymax=det_box[2] * img_height
                )
                for det_box in detection_boxes
            ]

            det_result = DetectionResult(
                det_bboxes=detection_bboxes,
                det_classes=detection_classes,
                det_scores=detection_scores,
                det_masks=detection_masks,
                det_time=inference_time,
                score_thr=score_thr
            )
            det_result.vis(image)

            if save_dir is not None:
                save_path = os.path.join(save_dir, os.path.basename(image_batch[ix])) if is_file(
                    image_batch[0]) else os.path.join(save_dir, str(ix) + ".jpg")
                imwrite(det_result.det_image, save_path)

            det_result_list.append(det_result)

        return det_result_list



from .colorspace import bgr2gray, gray2bgr, bgr2rgb, rgb2bgr, bgr2hsv, hsv2bgr
from .geometry import immerge,immix_up,imflip_lr, imflip_ud, imrotate, imcrop, impad, impad_to_multiple
from .normalize import imnormalize, imdenormalize
from .resize import imresize, imresize_like, imrescale

__all__ = [
    'bgr2gray', 'gray2bgr', 'bgr2rgb', 'rgb2bgr', 'bgr2hsv', 'hsv2bgr',
    'immerge','immix_up','imflip_lr', 'imflip_ud', 'imrotate', 'imcrop', 'impad', 'impad_to_multiple',
    'imnormalize', 'imdenormalize', 'imresize', 'imresize_like', 'imrescale'
]



# -*- coding: utf-8 -* -
import numpy as np
from io import BytesIO
from PIL import Image
import base64
import cv2
from .itis import is_np_array

def base64_to_np(b64_code):
    image = Image.open(BytesIO(base64.b64decode(b64_code)))
    img = np.array(image)
    return img

def np_to_base64(image_np):
    assert is_np_array(image_np)
    output_buffer = BytesIO()
    img = Image.fromarray(image_np.astype('uint8')).convert('RGB')
    img.save(output_buffer, format='JPEG')
    binary_data = output_buffer.getvalue()
    base64_data = base64.b64encode(binary_data)
    return base64_data

def get_mean_std(image_files):
    """
    获取图片均值方差
    :param image_files:
    :return:
    """
    ims = [np.array(cv2.imread(image_file)) for image_file in image_files]
    shape = ims[0].shape

    if len(shape) < 3:
        (mean, stddv) = cv2.meanStdDev(np.array(ims))
        return mean[0],stddv[0]
    else:
        im_mean = []
        im_std = []
        for dim in range(shape[2]):
            (mean, stddv) = cv2.meanStdDev(np.array(ims)[:, :, :, dim])
            im_mean.append(mean[0][0])
            im_std.append(stddv[0][0])
        return im_mean,im_std


# -*- coding: UTF-8 -*-
import numpy as np
from icv.data.core.polygon import PolygonInstance
from icv.data.core import FLIP_LEFT_RIGHT,FLIP_TOP_BOTTOM
from icv.data.core.mask_list import MaskList

class PolygonList(object):
    """
    This class handles PolygonInstances for all objects in the image
    """

    def __init__(self, polygons, size):
        """
        Arguments:
            polygons:
                a list of list of lists of numbers. The first
                level of the list correspond to individual instances,
                the second level to all the polygons that compose the
                object, and the third level to the polygon coordinates.
                OR
                a list of PolygonInstances.
                OR
                a PolygonList
            size: absolute image size
        """
        if isinstance(polygons, (list, tuple)):
            if len(polygons) == 0:
                polygons = [[[]]]
            if isinstance(polygons[0], (list, tuple)):
                assert isinstance(polygons[0][0], (list, tuple)), str(
                    type(polygons[0][0])
                )
            else:
                assert isinstance(polygons[0], PolygonInstance), str(type(polygons[0]))

        elif isinstance(polygons, PolygonList):
            size = polygons.size
            polygons = polygons.polygons

        else:
            RuntimeError(
                "Type of argument `polygons` is not allowed:%s" % (type(polygons))
            )

        assert isinstance(size, (list, tuple)), str(type(size))

        self.polygons = []
        for p in polygons:
            p = PolygonInstance(p, size)
            if len(p) > 0:
                self.polygons.append(p)

        self.size = tuple(size)

    def transpose(self, method):
        if method not in (FLIP_LEFT_RIGHT, FLIP_TOP_BOTTOM):
            raise NotImplementedError(
                "Only FLIP_LEFT_RIGHT and FLIP_TOP_BOTTOM implemented"
            )

        flipped_polygons = []
        for polygon in self.polygons:
            flipped_polygons.append(polygon.transpose(method))

        return PolygonList(flipped_polygons, size=self.size)

    def crop(self, box):
        w, h = box[2] - box[0], box[3] - box[1]
        cropped_polygons = []
        for polygon in self.polygons:
            cropped_polygons.append(polygon.crop(box))

        cropped_size = w, h
        return PolygonList(cropped_polygons, cropped_size)

    def resize(self, size):
        resized_polygons = []
        for polygon in self.polygons:
            resized_polygons.append(polygon.resize(size))

        resized_size = size
        return PolygonList(resized_polygons, resized_size)

    def to(self, *args, **kwargs):
        return self

    def convert_to_binarymask(self):
        if len(self) > 0:
            masks = np.stack([p.convert_to_binarymask() for p in self.polygons])
        else:
            size = self.size
            masks = np.empty([0, size[1], size[0]], dtype=np.uint8)

        return MaskList(masks)

    def __len__(self):
        return len(self.polygons)

    def __getitem__(self, item):
        if isinstance(item, int):
            selected_polygons = [self.polygons[item]]
        elif isinstance(item, slice):
            selected_polygons = self.polygons[item]
        else:
            # advanced indexing on a single dimension
            selected_polygons = []
            if isinstance(item,np.uint8):
                item = item.nonzero()
                item = item.squeeze(1) if item.numel() > 0 else item
                item = item.tolist()
            for i in item:
                selected_polygons.append(self.polygons[i])
        return PolygonList(selected_polygons, size=self.size)

    def __iter__(self):
        return iter(self.polygons)

    def __repr__(self):
        s = self.__class__.__name__ + "("
        s += "num_instances={}, ".format(len(self.polygons))
        s += "image_width={}, ".format(self.size[0])
        s += "image_height={})".format(self.size[1])
        return s


# -*- coding: UTF-8 -*-
import numpy as np
import copy
from .mask import Mask
from . import FLIP_TOP_BOTTOM,FLIP_LEFT_RIGHT

class PolygonInstance(object):
    """
    This class holds a set of polygons that represents a single instance
    of an object mask. The object can be represented as a set of
    polygons
    """

    def __init__(self, polygons, size):
        """
            Arguments:
                a list of lists of numbers.
                The first level refers to all the polygons that compose the
                object, and the second level to the polygon coordinates.
        """
        if isinstance(polygons, (list, tuple)):
            valid_polygons = []
            for p in polygons:
                # p = torch.as_tensor(p, dtype=torch.float32)
                if len(p) >= 6:  # 3 * 2 coordinates
                    valid_polygons.append(p)
            polygons = valid_polygons

        elif isinstance(polygons, PolygonInstance):
            polygons = copy.copy(polygons.polygons)

        else:
            RuntimeError(
                "Type of argument `polygons` is not allowed:%s" % (type(polygons))
            )

        """ This crashes the training way too many times...
        for p in polygons:
            assert p[::2].min() >= 0
            assert p[::2].max() < size[0]
            assert p[1::2].min() >= 0
            assert p[1::2].max() , size[1]
        """

        self.polygons = polygons
        self.size = tuple(size)

    def transpose(self, method):
        if method not in (FLIP_LEFT_RIGHT, FLIP_TOP_BOTTOM):
            raise NotImplementedError(
                "Only FLIP_LEFT_RIGHT and FLIP_TOP_BOTTOM implemented"
            )

        flipped_polygons = []
        width, height = self.size
        if method == FLIP_LEFT_RIGHT:
            dim = width
            idx = 0
        elif method == FLIP_TOP_BOTTOM:
            dim = height
            idx = 1

        for poly in self.polygons:
            p = poly.clone()
            TO_REMOVE = 1
            p[idx::2] = dim - poly[idx::2] - TO_REMOVE
            flipped_polygons.append(p)

        return PolygonInstance(flipped_polygons, size=self.size)

    def crop(self, box):
        assert isinstance(box, (list, tuple, np.ndarray)), str(type(box))
        box = list(box)

        # box is assumed to be xyxy
        current_width, current_height = self.size
        xmin, ymin, xmax, ymax = map(float, box)

        assert xmin <= xmax and ymin <= ymax, str(box)
        xmin = min(max(xmin, 0), current_width - 1)
        ymin = min(max(ymin, 0), current_height - 1)

        xmax = min(max(xmax, 0), current_width)
        ymax = min(max(ymax, 0), current_height)

        xmax = max(xmax, xmin + 1)
        ymax = max(ymax, ymin + 1)

        w, h = xmax - xmin, ymax - ymin

        cropped_polygons = []
        for poly in self.polygons:
            p = poly.clone()
            p[0::2] = p[0::2] - xmin  # .clamp(min=0, max=w)
            p[1::2] = p[1::2] - ymin  # .clamp(min=0, max=h)
            cropped_polygons.append(p)

        return PolygonInstance(cropped_polygons, size=(w, h))

    def resize(self, size):
        try:
            iter(size)
        except TypeError:
            assert isinstance(size, (int, float))
            size = size, size

        ratios = tuple(float(s) / float(s_orig) for s, s_orig in zip(size, self.size))

        if ratios[0] == ratios[1]:
            ratio = ratios[0]
            scaled_polys = [p * ratio for p in self.polygons]
            return PolygonInstance(scaled_polys, size)

        ratio_w, ratio_h = ratios
        scaled_polygons = []
        for poly in self.polygons:
            p = poly.clone()
            p[0::2] *= ratio_w
            p[1::2] *= ratio_h
            scaled_polygons.append(p)

        return PolygonInstance(scaled_polygons, size=size)

    def convert_to_binarymask(self):
        import pycocotools.mask as mask_utils

        width, height = self.size
        # formatting for COCO PythonAPI
        polygons = [p.numpy() for p in self.polygons]
        rles = mask_utils.frPyObjects(polygons, height, width)
        rle = mask_utils.merge(rles)
        mask = mask_utils.decode(rle)
        return Mask(mask)

    def __len__(self):
        return len(self.polygons)

    def __repr__(self):
        s = self.__class__.__name__ + "("
        s += "num_groups={}, ".format(len(self.polygons))
        s += "image_width={}, ".format(self.size[0])
        s += "image_height={}, ".format(self.size[1])
        return s






# -*- coding: UTF-8 -*-
from icv.image import imread,imshow,imwrite
from icv.utils import is_seq,is_file
from .meta import SampleMeta
from .anno_shape import AnnoShape
import os

class Anno(AnnoShape):
    def __init__(self,bbox=None,polys=None,mask=None,label=None,color=None,meta=None):
        super(Anno, self).__init__(bbox=bbox,polys=polys,mask=mask,label=label)
        self.color = color
        if is_seq(self.color):
            self.color = tuple(self.color)
        self.meta = meta

class Sample(object):
    def __init__(self,name,image,annos=None,meta=None):
        for anno in annos:
            assert isinstance(anno,Anno)

        assert meta is None or isinstance(meta,SampleMeta)
        assert is_file(image)

        self.name = name
        self.path = image
        _,self.ext = os.path.splitext(os.path.split(self.path)[1])
        self.image = imread(image)
        self.meta = meta
        self.annos = [] if annos is None else annos

    @staticmethod
    def init(name, image, **kwargs):
        sample = Sample(name, image, meta=SampleMeta(**kwargs))
        return sample

    @property
    def count(self):
        return len(self.annos)

    @property
    def shape(self):
        return self.image.shape

    @property
    def width(self):
        return self.shape[1]

    @property
    def height(self):
        return self.shape[0]

    @property
    def dim(self):
        return 1 if len(self.shape) == 2 else self.shape[-1]

    def statistics(self):
        sta = dict(count=self.count, cats=dict(), ratios=dict())
        for anno in self.annos:
            ratio = round(anno.bbox.ratio,1)
            if anno.label is not None:
                if anno.label not in sta["cats"]:
                    sta["cats"][anno.label] = 0
                sta["cats"][anno.label] += 1

                if anno.label not in sta["ratios"]:
                    sta["ratios"][anno.label] = dict()

                if ratio not in sta["ratios"][anno.label]:
                    sta["ratios"][anno.label][ratio] = 0

                sta["ratios"][anno.label][ratio] += 1

        return sta

    def vis(self,color=None,with_bbox=True, with_seg=True, is_show=False,save_path=None):
        image_drawed = self.image
        for anno in self.annos:
            _color = color if color is not None else anno.color
            _color = _color if _color is not None else "Orange"

            if with_bbox:
                image_drawed = anno.bbox.draw_on_image(image_drawed,color=_color)
            if with_seg and anno.seg_mode_mask:
                image_drawed = anno.mask.draw_on_image(image_drawed,color=_color)
            if with_seg and anno.seg_mode_polys:
                image_drawed = anno.polys.draw_on_image(image_drawed,color=_color)

        if is_show:
            imshow(image_drawed)

        if save_path is not None:
            imwrite(image_drawed,save_path)

        return image_drawed




# -*- coding: UTF-8 -*-
from icv.utils.itis import is_seq
import numpy as np
from .bbox import BBox
from collections import Iterable

class BBoxList(Iterable):
    def __init__(self,bbox_list):
        assert is_seq(bbox_list) or isinstance(bbox_list,BBoxList),"param bbox_list should be type of sequence or BBoxList"
        if isinstance(bbox_list,BBoxList):
            bbox_list = bbox_list.bbox_list
        bbox_list = list(bbox_list)

        for bbox_item in bbox_list:
            assert isinstance(bbox_item,BBox),"bbox should be type of BBox"

        self._labels = None
        self._bbox_list = bbox_list

    def select(self,ixs):
        if isinstance(ixs,np.ndarray):
            ixs = ixs.tolist()
        assert is_seq(ixs) and len(ixs) <= self.length
        self._bbox_list = [_ for ix,_ in enumerate(self._bbox_list) if ix in ixs]
        if self._labels is not None:
            self._labels = [_ for ix,_ in enumerate(self._labels) if ix in ixs]

    def __len__(self):
        return len(self.bbox_list)

    def __getitem__(self, item):
        assert item < self.length, "index out of the range."
        bbox = self.bbox_list[item]
        return bbox

    def __iter__(self):
        for bbox in self._bbox_list:
            yield bbox

    def tolist(self):
        return [[bbox.xmin,bbox.ymin,bbox.xmax,bbox.ymax] for bbox in self._bbox_list]

    def to_np_array(self):
        return np.array(self.tolist())

    @property
    def bbox_list(self):
        return self._bbox_list

    @property
    def length(self):
        return len(self)

    @property
    def labels(self):
        if self._labels:
            return self._labels

        self._labels = []
        for bbox in self.bbox_list:
            self._labels.append(bbox.lable)
        return self._labels

    @property
    def is_empty(self):
        return self.length == 0

    @property
    def xmin(self):
        return min([bbox.xmin for bbox in self.bbox_list])

    @property
    def ymin(self):
        return min([bbox.ymin for bbox in self.bbox_list])

    @property
    def xmax(self):
        return max([bbox.xmax for bbox in self.bbox_list])

    @property
    def ymax(self):
        return max([bbox.ymax for bbox in self.bbox_list])




# -*- coding: UTF-8 -*-
from ..utils import is_seq
import numpy as np
from ..image.vis import imshow_bboxes


class GroundTruth(object):
    def __init__(self, bboxes, classes, image):
        assert is_seq(bboxes), "param det_bboxes should be a sequence."
        self._bboxes = np.array(bboxes)
        self._classes = np.array(classes)
        self._image = image

    @property
    def bboxes(self):
        return self._bboxes

    @property
    def classes(self):
        return self._classes

    @property
    def image(self):
        return self._image

    @property
    def image_drawed(self):
        return self._image_drawed

    def vis(self, img, is_show=False, save_path=None):
        image_drawed = imshow_bboxes(img, self.bboxes, self.classes, is_show=is_show, save_path=save_path)
        self._image_drawed = image_drawed



# -*- coding: utf-8 -* -
import socket

def is_port_open(ip,port):
    s = socket.socket(socket.AF_INET,socket.SOCK_STREAM)
    try:
        s.connect((ip,int(port)))
        s.shutdown(2)
        return True
    except:
        return False

def collect_env_info():
    from torch.utils.collect_env import get_pretty_env_info
    env_str = get_pretty_env_info()
    return env_str




# -*- coding: UTF-8 -*-
import numpy as np
from icv.utils.itis import is_seq, is_np_array


def _transform_bboxes(bboxes):
    from icv.data.core import BBox, BBoxList
    if isinstance(bboxes, BBox):
        bboxes = BBoxList([bboxes])

    if isinstance(bboxes, BBoxList):
        bboxes = bboxes.to_np_array()

    if not is_np_array(bboxes):
        bboxes = np.array(bboxes)

    return bboxes


def bbox_clip(bboxes, img_shape):
    """Clip bboxes to fit the image shape.

    Args:
        bboxes (ndarray): Shape (..., 4*k)
        img_shape (tuple): (height, width) of the image.

    Returns:
        ndarray: Clipped bboxes.
    """
    bboxes = _transform_bboxes(bboxes)

    assert bboxes.shape[-1] % 4 == 0
    assert is_seq(img_shape) and len(img_shape) > 1
    clipped_bboxes = np.empty_like(bboxes, dtype=bboxes.dtype)
    clipped_bboxes[..., 0::2] = np.maximum(
        np.minimum(bboxes[..., 0::2], img_shape[1] - 1), 0)
    clipped_bboxes[..., 1::2] = np.maximum(
        np.minimum(bboxes[..., 1::2], img_shape[0] - 1), 0)
    return clipped_bboxes


def bbox_scaling(bboxes, scale, clip_shape=None):
    """Scaling bboxes w.r.t the box center.

    Args:
        bboxes (ndarray): Shape(..., 4).
        scale (float): Scaling factor.
        clip_shape (tuple, optional): If specified, bboxes that exceed the
            boundary will be clipped according to the given shape (h, w).

    Returns:
        ndarray: Scaled bboxes.
    """
    bboxes = _transform_bboxes(bboxes)
    assert bboxes.shape[-1] % 4 == 0

    if float(scale) == 1.0:
        scaled_bboxes = bboxes.copy()
    else:
        w = bboxes[..., 2] - bboxes[..., 0] + 1
        h = bboxes[..., 3] - bboxes[..., 1] + 1
        dw = (w * (scale - 1)) * 0.5
        dh = (h * (scale - 1)) * 0.5
        scaled_bboxes = bboxes + np.stack((-dw, -dh, dw, dh), axis=-1)
    if clip_shape is not None:
        return bbox_clip(scaled_bboxes, clip_shape)
    else:
        return scaled_bboxes


def bbox_extend(bboxes, pad, clip_shape=None):
    """
    Extend bboxes to new position
    :param bboxes (ndarray): Shape(..., 4).
    :param pad: (int or 2 value list or 4 value list order from top、right、bottom、left)
    :param clip_shape(tuple, optional): If specified, bboxes that exceed the
            boundary will be clipped according to the given shape (h, w).
    :return:
        ndarray: Extended bboxes.
    """
    if isinstance(pad, int):
        t, r, b, l = (pad,) * 4
    elif is_seq(pad):
        if len(pad) == 1:
            t, r, b, l = (pad[0],) * 4
        elif len(pad) == 2:
            t, b = (pad[0],) * 2
            r, l = (pad[1],) * 2
        elif len(pad) >= 4:
            t, r, b, l = pad[0], pad[1], pad[2], pad[3]
        else:
            raise ValueError("pad should be one of [int(top|right|bottom|left),"
                             "tuple(2,top|bottom,right|left),tuple(4,top,right,bottom,left)]")
    else:
        raise ValueError("pad should be one of [int(top|right|bottom|left),"
                         "tuple(2,top|bottom,right|left),tuple(4,top,right,bottom,left)]")

    bboxes = _transform_bboxes(bboxes)
    assert bboxes.shape[-1] % 4 == 0

    extended_bboxes = bboxes.copy()
    extended_bboxes[..., 0] -= l
    extended_bboxes[..., 1] -= t
    extended_bboxes[..., 2] += r
    extended_bboxes[..., 3] += b

    if clip_shape is not None:
        return bbox_clip(extended_bboxes, clip_shape)
    else:
        return extended_bboxes



from .itis import is_str, is_list, is_dir, is_seq, is_empty, is_file, is_py3, IS_PY3, is_seq_equal, is_valid_url, \
    is_np_array, is_float_array
from .misc import check_file_exist, concat_list
from .path import mkfile, mkdir, mkdir_force, reset_dir, fcopy, list_from_file, list_to_file
from .time import Time
from .timer import Timer
from .anno import load_voc_anno, save_voc_anno, make_empty_coco_anno, make_empty_voc_anno
from .opencv import USE_OPENCV2
from .label_map_util import labelmap_to_category_index, labelmap_to_categories
from .checkpoint import ckpt_load, ckpt_save
from .config import Config
from .image import base64_to_np, np_to_base64, get_mean_std
from .codec import json_encode, json_decode, encode_to_file, decode_from_file
from .easy_dict import EasyDict
from .require import requires_package, requires_executable
from .http import request, do_post, do_get, do_put, do_delete, do_options, simple, file_server

__all__ = [
    'is_str', 'is_list', 'is_dir', 'is_seq', 'is_empty', 'is_file', 'is_py3', 'IS_PY3', 'is_seq_equal', 'is_valid_url',
    'is_np_array', 'is_float_array',
    'check_file_exist', 'concat_list', 'mkfile', 'mkdir', 'mkdir_force', 'reset_dir', 'fcopy', 'list_from_file',
    'list_to_file',
    'Time',
    'Timer', 'load_voc_anno', 'save_voc_anno', 'make_empty_coco_anno', 'make_empty_voc_anno', 'USE_OPENCV2',
    'labelmap_to_category_index',
    'labelmap_to_categories', 'requires_package', 'requires_executable', 'json_encode', 'json_decode', 'encode_to_file',
    'decode_from_file',
    'ckpt_load', 'ckpt_save', 'Config', 'base64_to_np', 'np_to_base64', 'EasyDict',
    'request', 'do_post', 'do_get', 'do_put', 'do_delete', 'do_options', 'simple', 'file_server'
]



# -*- coding: UTF-8 -*-
import os
import itertools

def check_file_exist(filename, msg_tmpl='file "{}" does not exist'):
    if not os.path.isfile(filename):
        raise FileNotFoundError(msg_tmpl.format(filename))

def concat_list(in_list):
    """Concatenate a list of list into a single list.

    Args:
        in_list (list): The list of list to be merged.

    Returns:
        list: The concatenated flat list.
    """
    return list(itertools.chain(*in_list))


# -*- coding: UTF-8 -*-

class HttpParseException(Exception):
    # Exception raised by http request parser
    def __init__(self,msg,code=500):
        super(HttpParseException, self).__init__(msg)
        self.msg = msg
        self.code = code



# -*- coding: UTF-8 -*-
try:
    import maskrcnn_benchmark
except ImportError as e:
    raise Exception("You should install maskrcnn_benchmark first " \
          "(https://github.com/facebookresearch/maskrcnn-benchmark/blob/master/INSTALL.md) . " \
          "Import maskrcnn_benchmark Error!")

from .detector import Detector
from .result import DetectionResult
from ..image import imread,imwrite,imshow,imresize
import torch
import os
import numpy as np
from ..utils import Timer,is_file
from torchvision import transforms as T
from ..data.core import BBox
from maskrcnn_benchmark.modeling.detector import build_detection_model
from maskrcnn_benchmark.utils.checkpoint import DetectronCheckpointer
from maskrcnn_benchmark.structures.image_list import to_image_list
from maskrcnn_benchmark.modeling.roi_heads.mask_head.inference import Masker
from maskrcnn_benchmark.structures.bounding_box import BoxList
from maskrcnn_benchmark.config import cfg

class MbDetector(Detector):
    def __init__(self,model_path,config_file,categories,show_mask_heatmaps=False,iou_thr=0.5,score_thr=0.5,device=None):
        super(MbDetector, self).__init__(categories=categories, iou_thr=iou_thr, score_thr=score_thr,device=device)

        assert is_file(model_path),"model path does not exist!"
        assert is_file(config_file),"config path does not exist!"

        self.model_path = model_path
        self.config_file = config_file
        cfg.merge_from_file(config_file)
        self.cfg = cfg.clone()
        self.cpu_device = torch.device("cpu")

        mask_threshold = -1 if show_mask_heatmaps else 0.5
        self.masker = Masker(threshold=mask_threshold, padding=1)
        self.show_mask_heatmaps = show_mask_heatmaps

        self._build_detector()

    def _build_detector(self):
        self.model = build_detection_model(self.cfg)
        self.model.eval()
        self.model.to(self.device)

        checkpointer = DetectronCheckpointer(self.cfg, self.model)
        _ = checkpointer.load(self.model_path)

        self.transforms = self._build_transform()

    def _build_transform(self):
        """
        Creates a basic transformation that was used to train the models
        """
        cfg = self.cfg
        if cfg.INPUT.TO_BGR255:
            to_bgr_transform = T.Lambda(lambda x: x * 255)
        else:
            to_bgr_transform = T.Lambda(lambda x: x[[2, 1, 0]])

        normalize_transform = T.Normalize(
            mean=cfg.INPUT.PIXEL_MEAN, std=cfg.INPUT.PIXEL_STD
        )

        transform = T.Compose(
            [
                T.ToPILImage(),
                T.ToTensor(),
                to_bgr_transform,
                normalize_transform,
            ]
        )
        return transform

    def _compute_prediction(self, original_image):
        """
        Arguments:
            original_image (np.ndarray): an image as returned by OpenCV

        Returns:
            prediction (BoxList): the detected objects. Additional information
                of the detection properties can be found in the fields of
                the BoxList via `prediction.fields()`
        """
        # apply pre-processing to image
        image = self.transforms(original_image)
        # convert to an ImageList, padded so that it is divisible by
        # cfg.DATALOADER.SIZE_DIVISIBILITY
        image_list = to_image_list(image, self.cfg.DATALOADER.SIZE_DIVISIBILITY)
        image_list = image_list.to(self.device)
        # compute predictions
        with torch.no_grad():
            predictions = self.model(image_list)
        predictions = [o.to(self.cpu_device) for o in predictions]

        # always single image is passed at a time
        prediction = predictions[0]

        # reshape prediction (a BoxList) into the original image size
        height, width = original_image.shape[:-1]
        prediction = prediction.resize((width, height))

        if prediction.has_field("mask"):
            # if we have masks, paste the masks in the right position
            # in the image, as defined by the bounding boxes
            masks = prediction.get_field("mask")
            # always single image is passed at a time
            masks = self.masker([masks], [prediction])[0]
            prediction.add_field("mask", masks)
        return prediction

    def _compute_prediction_on_batch(self,original_image_list):
        """
        batch inference
        :param original_image_list: image list and each image as returned by OpenCV
        :return: batch inference result
        """
        image_list = [self.transforms(each_image) for each_image in original_image_list]
        image_list = to_image_list(image_list, self.cfg.DATALOADER.SIZE_DIVISIBILITY)
        image_list = image_list.to(self.device)
        # compute predictions
        with torch.no_grad():
            predictions = self.model(image_list)

        def post_process(original_image,prediction):
            prediction.to(self.cpu_device)
            height, width = original_image.shape[:-1]
            prediction = prediction.resize((width, height))
            if prediction.has_field("mask"):
                # if we have masks, paste the masks in the right position
                # in the image, as defined by the bounding boxes
                masks = prediction.get_field("mask")
                # always single image is passed at a time
                masks = self.masker([masks], [prediction])[0]
                prediction.add_field("mask", masks)

            return prediction

        predictions = [post_process(original_image, pred) for original_image, pred in
                  list(zip(original_image_list, predictions))]
        return predictions

    def _build_result(self, inference_result, inference_time=0, score_thr=-1):
        assert isinstance(inference_result,BoxList),"inference result should be type of BoxList!"
        classes_result = inference_result.get_field("labels")
        scores_result = inference_result.get_field("scores")
        bbox_result = inference_result.bbox

        segm_result = None
        if inference_result.has_field("mask"):
            segm_result = inference_result.get_field("mask")

        detection_classes = []
        detection_scores = []
        detection_bboxes = []
        detection_segm = []
        for ix in range(bbox_result.shape[0]):
            xmin,ymin,xmax,ymax = bbox_result[ix,:]
            if xmin >= xmax or ymin >= ymax:
                continue
            detection_classes.append(classes_result[ix])
            detection_scores.append(scores_result[ix])
            detection_bboxes.append(BBox(xmin,ymin,xmax,ymax))
            if segm_result is not None:
                detection_segm.append(segm_result[ix,...])

        detection_segm = None if len(detection_segm) == 0 else detection_segm

        score_thr = score_thr if score_thr >= 0 else self.score_thr
        det_result = DetectionResult(
            det_bboxes=detection_bboxes,
            det_classes=np.array(detection_classes),
            det_scores=np.array(detection_scores),
            det_masks=np.array(detection_segm),
            det_time=inference_time,
            categories=self.categories,
            score_thr=score_thr
        )

        return det_result

    def inference(self, image, is_show=False, save_path=None, score_thr=-1):
        image_np = imread(image)
        timer = Timer()
        result = self._compute_prediction(image_np)
        inference_time = timer.since_start()

        score_thr = score_thr if score_thr >= 0 else self.score_thr
        det_result = self._build_result(result, inference_time, score_thr)
        det_result.vis(image)

        if is_show:
            imshow(det_result.det_image)

        if save_path is not None:
            imwrite(det_result.det_image, save_path)

        return det_result

    def inference_batch(self, image_batch, save_dir=None, resize=None, score_thr=-1):
        image_np_list = [imread(img) for img in image_batch]
        if resize:
            image_np_list = [imresize(img_np,resize) for img_np in image_np_list]

        if len(set([img_np.shape for img_np in image_np_list])) != 1:
            return [self.inference(img_np) for img_np in image_np_list]

        timer = Timer()
        predictions = self._compute_prediction_on_batch(image_np_list)
        inference_time = timer.since_start()
        score_thr = score_thr if score_thr >= 0 else self.score_thr

        det_result_list = []
        for ix,result in enumerate(predictions):
            det_result = self._build_result(result,inference_time,score_thr)
            det_result.vis(image_np_list[ix])

            if save_dir is not None:
                save_path = os.path.join(save_dir,os.path.basename(image_batch[ix])) if is_file(image_batch[0]) else os.path.join(save_dir,str(ix)+".jpg")
                imwrite(det_result.det_image, save_path)

            det_result_list.append(det_result)

        return det_result_list


# -*- coding: UTF-8 -*-
import numpy as np
from skimage import measure
from icv.utils import EasyDict as edict
from icv.utils import is_seq, is_np_array,is_file
from .polys import Polygon
from icv.image.io import imread,imwrite
from icv.image.vis import imdraw_mask


class Mask(object):
    """
    Binary Mask
    """

    def __init__(self, mask, label=None, **kwargs):
        assert is_seq(mask) or is_np_array(mask)
        self.mask = np.array(mask).astype(np.uint8)
        assert len(self.mask.shape) == 2
        self.size = tuple(self.mask.shape)

        self.mask[self.mask != 0] = 1
        self.label = label

        self.polygon = None
        self.bbox = None

        self.fields = edict()
        for k in kwargs:
            self.add_field(k, kwargs[k])

        self.label = label
        if label is not None:
            self.add_field("label", self.label)

    @staticmethod
    def init_from(mask, label=None):
        if isinstance(mask, Mask):
            return mask.deepcopy(label=label)
        return Mask(mask, label=label)

    def add_field(self, field, field_data):
        self.fields[field] = field_data

    def get_field(self, field):
        return self.fields[field]

    def has_field(self, field):
        return field in self.fields

    def fields(self):
        return list(self.fields.keys())

    def find_contours(self):
        contours = measure.find_contours(self.mask, 0.5)

        reshaped_contour = []
        for contour in contours:
            contour = np.flip(contour, axis=1)
            s = contour.ravel().tolist()
            assert len(s) % 2 == 0
            s = [(s[i], s[i + 1]) for i in range(len(s)) if i % 2 == 0]
            reshaped_contour.append(s)

        return reshaped_contour

        # mask = cv2.UMat(self.mask)
        # contour, hierarchy = cv2.findContours(
        #     mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_TC89_L1
        # )
        #
        # reshaped_contour = []
        # for entity in contour:
        #     assert len(entity.shape) == 3
        #     assert entity.shape[1] == 1, "Hierarchical contours are not allowed"
        #     reshaped_contour.append(entity.reshape(-1).tolist())
        #
        # return reshaped_contour

    def to_ploygons(self):
        if self.polygon is not None:
            return self.polygon
        contour = self.find_contours()
        if len(contour) > 0:
            self.polygon = Polygon.init_from(contour[0], self.label)
        else:
            self.polygon = Polygon.init_from([[0, 0]], self.label)
        return self.polygon

    def to_bounding_box(self):
        if self.bbox is not None:
            return self.bbox
        self.bbox = self.to_ploygons().to_bounding_box()
        return self.bbox

    def copy(self, mask=None, label=None, **kwargs):
        """
        Create a shallow copy of the Polygon object.
        Parameters
        ----------
        exterior : list of imgaug.Keypoint or list of tuple or (N,2) ndarray, optional
            List of points defining the polygon. See :func:`imgaug.Polygon.__init__` for details.
        label : None or str, optional
            If not None, then the label of the copied object will be set to this value.
        Returns
        -------
        imgaug.Polygon
            Shallow copy.
        """
        return self.deepcopy(mask=mask, label=label, **kwargs)

    def deepcopy(self, mask=None, label=None, **kwargs):
        """
        Create a deep copy of the Polygon object.
        Parameters
        ----------
        exterior : list of Keypoint or list of tuple or (N,2) ndarray, optional
            List of points defining the polygon. See `imgaug.Polygon.__init__` for details.
        label : None or str
            If not None, then the label of the copied object will be set to this value.
        Returns
        -------
        imgaug.Polygon
            Deep copy.
        """
        return Mask(
            mask=np.copy(self.mask) if mask is None else mask,
            label=self.label if label is None else label,
            **kwargs
        )

    def draw_on_image(self, image, color=(0, 255, 0), alpha=0.6,
                      copy=True, raise_if_out_of_image=False):
        assert image.shape[:2] == self.mask.shape

        bbox = self.to_bounding_box()
        if raise_if_out_of_image and bbox.is_out_of_image(image):
            raise Exception(
                "Cannot draw mask xmin = %.8f, ymin = %.8f, xmax = %.8f, ymax = %.8f on image with shape %s." % (
                    bbox.xmin, bbox.ymin, bbox.xmax, bbox.ymax, image.shape))

        result = np.copy(image) if copy else image

        if isinstance(color, (tuple, list)):
            color = np.uint8(color)

        return imdraw_mask(image=result, mask=self.mask, color=color, alpha=alpha)

    def save(self, dist_path, id=1):
        m = self.mask[::]
        m[np.where(m != 0)] = id
        if is_file(dist_path):
            m0 = imread(dist_path,0)
            m = m + m0
            m[np.where(m>max(id,np.max(m0)))] = id

        imwrite(m, dist_path)



# -*- coding: UTF-8 -*-
from ..utils import is_seq
from ..data.core import BBoxList
from ..image import imshow_bboxes
import numpy as np

class DetectionResult(object):
    def __init__(self,det_bboxes,det_classes,det_scores,det_masks=None,det_time=0,det_image=None,categories=None,score_thr=0):
        assert is_seq(det_bboxes), "param det_bboxes should be a sequence."
        self._score_thr = max(score_thr,0)
        self._det_scores = np.array(det_scores)

        self._det_bboxes = BBoxList(np.array(det_bboxes).tolist())
        self._det_classes = np.array(det_classes)
        self._det_masks = None if det_masks is None else np.array(det_masks)
        self._det_time = det_time
        self._det_image = det_image
        self.categories = categories
        self._det_labels = None
        if self.categories is not None:
            self._det_labels = [self.categories[_-1] for _ in self._det_classes]

        if self._score_thr >= 0:
            self.select_top_predictions()

    def select_top_predictions(self):
        filter_ids = np.where(np.array(self.det_scores) >= self._score_thr)
        self._det_scores = self._det_scores[filter_ids]
        self._det_bboxes.select(filter_ids[0])
        self._det_classes = self._det_classes[filter_ids]
        # TODO: 这里的过滤失效？？
        if self._det_masks is not None and self._det_masks.shape[0] > 0:
            self._det_masks = self._det_masks[filter_ids]

        if self.categories is not None:
            self._det_labels = [self.categories[_-1] for _ in self._det_classes]

    @property
    def det_bboxes(self):
        return self._det_bboxes

    @property
    def det_classes(self):
        return self._det_classes

    @property
    def det_labels(self):
        return self._det_labels

    @property
    def det_scores(self):
        return self._det_scores

    @property
    def det_masks(self):
        return self._det_masks

    @property
    def det_time(self):
        return self._det_time

    @property
    def det_image(self):
        return self._det_image

    def topk(self,k=1):
        if len(self) == 0:
            return [(-1,-1)]

        assert k <= len(self),"param k should smaller than bbox count."

        topk_idx = np.argsort(self.det_scores)[-k:]
        topk_class = self.det_classes[topk_idx][::-1]
        topk_score = self.det_scores[topk_idx][::-1]

        _topk = list(zip(topk_class.tolist(),topk_score.tolist()))
        return _topk

    @property
    def length(self):
        return len(self)

    def __len__(self):
        return len(self.det_bboxes)

    def to_dict(self):
        return dict(
            bboxes=self.det_bboxes.tolist(),
            classes=self.det_classes,
            scores=self.det_scores,
            time=self.det_time,
            image=self.det_image
        )

    # TODO: add image return
    def to_json(self):
        return dict(
            bboxes=self.det_bboxes.tolist(),
            classes=self.det_classes.tolist() if isinstance(self.det_classes,np.ndarray) else self.det_classes,
            scores=self.det_scores.tolist() if isinstance(self.det_classes,np.ndarray) else self.det_classes,
            time=self.det_time
        )

    def vis(self,img):
        image_drawed = imshow_bboxes(img,
                                     self.det_bboxes,
                                     classes=self.categories,
                                     labels=self.det_labels,
                                     scores=self.det_scores,
                                     masks=self.det_masks
                                     )
        self._det_image = image_drawed
        return image_drawed



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: object_detection/protos/string_int_label_map.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='object_detection/protos/string_int_label_map.proto',
  package='object_detection.protos',
  syntax='proto2',
  serialized_options=None,
  serialized_pb=_b('\n2object_detection/protos/string_int_label_map.proto\x12\x17object_detection.protos\"G\n\x15StringIntLabelMapItem\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\n\n\x02id\x18\x02 \x01(\x05\x12\x14\n\x0c\x64isplay_name\x18\x03 \x01(\t\"Q\n\x11StringIntLabelMap\x12<\n\x04item\x18\x01 \x03(\x0b\x32..object_detection.protos.StringIntLabelMapItem')
)




_STRINGINTLABELMAPITEM = _descriptor.Descriptor(
  name='StringIntLabelMapItem',
  full_name='object_detection.protos.StringIntLabelMapItem',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='object_detection.protos.StringIntLabelMapItem.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='id', full_name='object_detection.protos.StringIntLabelMapItem.id', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='display_name', full_name='object_detection.protos.StringIntLabelMapItem.display_name', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=79,
  serialized_end=150,
)


_STRINGINTLABELMAP = _descriptor.Descriptor(
  name='StringIntLabelMap',
  full_name='object_detection.protos.StringIntLabelMap',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='item', full_name='object_detection.protos.StringIntLabelMap.item', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=152,
  serialized_end=233,
)

_STRINGINTLABELMAP.fields_by_name['item'].message_type = _STRINGINTLABELMAPITEM
DESCRIPTOR.message_types_by_name['StringIntLabelMapItem'] = _STRINGINTLABELMAPITEM
DESCRIPTOR.message_types_by_name['StringIntLabelMap'] = _STRINGINTLABELMAP
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

StringIntLabelMapItem = _reflection.GeneratedProtocolMessageType('StringIntLabelMapItem', (_message.Message,), dict(
  DESCRIPTOR = _STRINGINTLABELMAPITEM,
  __module__ = 'object_detection.protos.string_int_label_map_pb2'
  # @@protoc_insertion_point(class_scope:object_detection.protos.StringIntLabelMapItem)
  ))
_sym_db.RegisterMessage(StringIntLabelMapItem)

StringIntLabelMap = _reflection.GeneratedProtocolMessageType('StringIntLabelMap', (_message.Message,), dict(
  DESCRIPTOR = _STRINGINTLABELMAP,
  __module__ = 'object_detection.protos.string_int_label_map_pb2'
  # @@protoc_insertion_point(class_scope:object_detection.protos.StringIntLabelMap)
  ))
_sym_db.RegisterMessage(StringIntLabelMap)


# @@protoc_insertion_point(module_scope)



# -*- coding: UTF-8 -*-
from ..voc import Voc
from ..coco import Coco
from ..core.sample import Sample
from icv.utils import reset_dir, mkfile
import os


class CocoConverter(object):
    def __init__(self, coco):
        assert isinstance(coco, Coco)
        self.coco = coco

    def to_voc(self, voc_root, split=None,reset=False):
        split = split if split is not None else self.coco.split
        if reset:
            reset_dir(voc_root)
        anno_path, image_path, imgset_path, imgset_seg_path, \
        seg_class_image_path, seg_object_image_path = Voc.reset_dir(voc_root)

        setfile = os.path.join(imgset_seg_path, "%s.txt" % split) if self.coco.is_seg_mode else os.path.join(
            imgset_path, "%s.txt" % split)
        mkfile(setfile)

        voc = Voc(
            voc_root,
            split,
            keep_no_anno_image=self.coco.keep_no_anno_image,
            mode="detect" if not self.coco.is_seg_mode else "segment",
            categories=self.coco.categories,
            one_index=self.coco.one_index,
        )

        for sample in self.coco.get_samples():
            assert isinstance(sample, Sample)
            voc.sample_db[sample.name] = sample
            voc.ids.append(sample.name)

        voc.save(voc_root, reset_dir=True, split=split)
        voc.init()
        return voc



# -*- coding: UTF-8 -*-
import os

def make_video(images, outpath, fps=5, size=None,
               is_color=True, format="XVID"):
    """
    Create a video from a list of images.

    @param      outpath      output video
    @param      images      list of images to use in the video
    @param      fps         frame per second
    @param      size        size of each frame
    @param      is_color    color
    @param      format      see http://www.fourcc.org/codecs.php
    @return                 see http://opencv-python-tutroals.readthedocs.org/en/latest/py_tutorials/py_gui/py_video_display/py_video_display.html

    The function relies on http://opencv-python-tutroals.readthedocs.org/en/latest/.
    By default, the video will have the size of the first image.
    It will resize every image to this size before adding them to the video.
    """
    from cv2 import VideoWriter, VideoWriter_fourcc, imread, resize
    fourcc = VideoWriter_fourcc(*format)
    vid = None
    for image in images:
        if not os.path.exists(image):
            raise FileNotFoundError(image)
        img = imread(image)
        if vid is None:
            if size is None:
                size = img.shape[1], img.shape[0]
            vid = VideoWriter(outpath, fourcc, float(fps), size, is_color)
        if size[0] != img.shape[1] and size[1] != img.shape[0]:
            img = resize(img, size)
        vid.write(img)
    vid.release()
    return vid



# -*- coding: UTF-8 -*-
import os
import numpy as np

def is_str(input):
    return isinstance(input, str)


def is_list(input):
    return isinstance(input, list)


def is_seq(input):
    return isinstance(input, tuple) or isinstance(input, list)


def is_seq_equal(seq1, seq2, strict=False):
    """
    判断序列是否相等
    :param seq1: 序列1
    :param seq2: 序列2
    :param strict: 是否严格模式，即顺序也要匹配
    :return:
    """
    if not is_seq(seq1) or not is_seq(seq2) or len(seq1) != len(seq2):
        return False

    if strict:
        for i1, i2 in list(zip(seq1, seq2)):
            if i1 != i2:
                return False
        return True

    for i in seq1:
        if i not in seq2:
            return False
    return True


def is_np_array(input):
    return isinstance(input, np.ndarray)


def is_float_array(input):
    return is_np_array(input) and issubclass(input.dtype.type, np.floating)


def is_file(input):
    return is_str(input) and os.path.isfile(input) and os.path.exists(input)


def is_dir(input):
    return is_str(input) and os.path.isdir(input) and os.path.exists(input)


def is_path(input):
    return is_file(input) or is_dir(input)


def is_empty(input):
    if is_str(input):
        return input == ""
    if is_seq(input):
        return len(input) == 0
    if is_file(input):
        return float(os.path.getsize(input)) > 0
    if is_dir(input):
        from glob import glob
        return len(glob(os.path.join(input,"**","*"),recursive=True)) > 0

    return input is None


def is_valid_url(url):
    import re
    if re.match(r'^https?:/{2}\w.+$', url):
        return True
    return False


def is_py3():
    import sys
    if sys.version_info > (3, 0):
        return True
    return False


IS_PY3 = is_py3()



# -*- coding: UTF-8 -*-

class IcvHttpResponse(object):
    content=None
    status_code = 200
    error = None
    headers = {}

    def __init__(self, content=None,status_code=200):
        self.content = content
        self.status_code = status_code

    @property
    def ok(self):
        return self.status_code == 200

    def __repr__(self):
        s = ""
        s += "===content===\n" + str(self.content)
        s += "===status code===\n" + str(self.status_code)
        s += "===error===\n" + str(self.error)
        s += "===headers===\n" + str(self.headers)
        return s



# -*- coding: UTF-8 -*-
try:
    # Python 2.x
    from SocketServer import ThreadingMixIn
    from SimpleHTTPServer import SimpleHTTPRequestHandler
    from BaseHTTPServer import HTTPServer
except ImportError:
    # Python 3.x
    from socketserver import ThreadingMixIn
    from http.server import SimpleHTTPRequestHandler, HTTPServer

class ThreadingSimpleServer(ThreadingMixIn, HTTPServer):
    pass

class IcvServer(object):
    def __init__(self,host="0.0.0.0",port=9527):
        self.host = host
        self.port = port
        self.server = None
        self.handler = IcvServer

    def run(self,handler_class):
        self.server = ThreadingSimpleServer((self.host, self.port), handler_class)
        self.server.serve_forever()




# -*- coding: UTF-8 -*-

class HttpMethod(object):
    POST = "POST"
    GET = "GET"
    PUT = "PUT"
    DELETE = "DELETE"
    HEAD = "HEAD"
    CONNECT = "CONNECT"
    OPTIONS = "OPTIONS"
    TRACE = "TRACE"
    PATCH = "PATCH"

    methods = [
            POST,
            GET,
            PUT,
            DELETE,
            HEAD,
            CONNECT,
            OPTIONS,
            TRACE,
            PATCH,
    ]



# -*- coding: UTF-8 -*-
from icv.data.core.mask_list import MaskList
from icv.data.core.polygon_list import PolygonList

class SegmentationMask(object):

    """
    This class stores the segmentations for all objects in the image.
    It wraps BinaryMaskList and PolygonList conveniently.
    """

    def __init__(self, instances, size, mode="poly"):
        """
        Arguments:
            instances: two types
                (1) polygon
                (2) binary mask
            size: (width, height)
            mode: 'poly', 'mask'. if mode is 'mask', convert mask of any format to binary mask
        """

        assert isinstance(size, (list, tuple))
        assert len(size) == 2

        assert isinstance(size[0], (int, float))
        assert isinstance(size[1], (int, float))

        if mode == "poly":
            self.instances = PolygonList(instances, size)
        elif mode == "mask":
            self.instances = MaskList(instances)
        else:
            raise NotImplementedError("Unknown mode: %s" % str(mode))

        self.mode = mode
        self.size = tuple(size)

    # TODO: 尚未实现
    def transpose(self, method):
        flipped_instances = self.instances.transpose(method)
        return SegmentationMask(flipped_instances, self.size, self.mode)

    # TODO：尚未实现
    def crop(self, box):
        cropped_instances = self.instances.crop(box)
        cropped_size = cropped_instances.size
        return SegmentationMask(cropped_instances, cropped_size, self.mode)

    def resize(self, size, *args, **kwargs):
        resized_instances = self.instances.resize(size)
        resized_size = size
        return SegmentationMask(resized_instances, resized_size, self.mode)

    def to(self, *args, **kwargs):
        return self

    def convert(self, mode):
        if mode == self.mode:
            return self

        if mode == "poly":
            converted_instances = self.instances.convert_to_polygon()
        elif mode == "mask":
            converted_instances = self.instances.convert_to_binarymask()
        else:
            raise NotImplementedError("Unknown mode: %s" % str(mode))

        return SegmentationMask(converted_instances, self.size, mode)

    def get_mask_tensor(self):
        instances = self.instances
        if self.mode == "poly":
            instances = instances.convert_to_binarymask()
        # If there is only 1 instance
        return instances.masks.squeeze(0)

    def __len__(self):
        return len(self.instances)

    def __getitem__(self, item):
        selected_instances = self.instances.__getitem__(item)
        return SegmentationMask(selected_instances, self.size, self.mode)

    def __iter__(self):
        self.iter_idx = 0
        return self

    def __next__(self):
        if self.iter_idx < self.__len__():
            next_segmentation = self.__getitem__(self.iter_idx)
            self.iter_idx += 1
            return next_segmentation
        raise StopIteration()

    next = __next__  # Python 2 compatibility

    def __repr__(self):
        s = self.__class__.__name__ + "("
        s += "num_instances={}, ".format(len(self.instances))
        s += "image_width={}, ".format(self.size[0])
        s += "image_height={}, ".format(self.size[1])
        s += "mode={})".format(self.mode)
        return s



# -*- coding: UTF-8 -*-
import functools
import subprocess
from importlib import import_module

def check_prerequisites(
        prerequisites,
        checker,
        msg_tmpl='Prerequisites "{}" are required in method "{}" but not '
        'found, please install them first.'):
    """A decorator factory to check if prerequisites are satisfied.
    Args:
        prerequisites (str of list[str]): Prerequisites to be checked.
        checker (callable): The checker method that returns True if a
            prerequisite is meet, False otherwise.
        msg_tmpl (str): The message template with two variables.
    Returns:
        decorator: A specific decorator.
    """

    def wrap(func):

        @functools.wraps(func)
        def wrapped_func(*args, **kwargs):
            requirements = [prerequisites] if isinstance(
                prerequisites, str) else prerequisites
            missing = []
            for item in requirements:
                if not checker(item):
                    missing.append(item)
            if missing:
                print(msg_tmpl.format(', '.join(missing), func.__name__))
                raise RuntimeError('Prerequisites not meet.')
            else:
                return func(*args, **kwargs)

        return wrapped_func

    return wrap


def _check_py_package(package):
    try:
        import_module(package)
    except ImportError:
        return False
    else:
        return True


def _check_executable(cmd):
    if subprocess.call('which {}'.format(cmd), shell=True) != 0:
        return False
    else:
        return True


def requires_package(prerequisites):
    """A decorator to check if some python packages are installed.
    Example:
        >>> @requires_package('numpy')
        >>> func(arg1, args):
        >>>     return numpy.zeros(1)
        array([0.])
        >>> @requires_package(['numpy', 'non_package'])
        >>> func(arg1, args):
        >>>     return numpy.zeros(1)
        ImportError
    """
    return check_prerequisites(prerequisites, checker=_check_py_package)


def requires_executable(prerequisites):
    """A decorator to check if some executable files are installed.
    Example:
        >>> @requires_executable('ffmpeg')
        >>> func(arg1, args):
        >>>     print(1)
        1
    """
    return check_prerequisites(prerequisites, checker=_check_executable)






# -*- coding: UTF-8 -*-
import numpy as np

def bbox_overlaps(bboxes1, bboxes2, mode='iou'):
    """Calculate the ious between each bbox of bboxes1 and bboxes2.
    Args:
        bboxes1(ndarray): shape (n, 4)
        bboxes2(ndarray): shape (k, 4)
        mode(str): iou (intersection over union) or iof (intersection
            over foreground)
    Returns:
        ious(ndarray): shape (n, k)
    """

    assert mode in ['iou', 'iof']

    bboxes1 = bboxes1.astype(np.float32)
    bboxes2 = bboxes2.astype(np.float32)
    rows = bboxes1.shape[0]
    cols = bboxes2.shape[0]
    ious = np.zeros((rows, cols), dtype=np.float32)
    if rows * cols == 0:
        return ious
    exchange = False
    if bboxes1.shape[0] > bboxes2.shape[0]:
        bboxes1, bboxes2 = bboxes2, bboxes1
        ious = np.zeros((cols, rows), dtype=np.float32)
        exchange = True
    area1 = (bboxes1[:, 2] - bboxes1[:, 0] + 1) * (
        bboxes1[:, 3] - bboxes1[:, 1] + 1)
    area2 = (bboxes2[:, 2] - bboxes2[:, 0] + 1) * (
        bboxes2[:, 3] - bboxes2[:, 1] + 1)
    for i in range(bboxes1.shape[0]):
        x_start = np.maximum(bboxes1[i, 0], bboxes2[:, 0])
        y_start = np.maximum(bboxes1[i, 1], bboxes2[:, 1])
        x_end = np.minimum(bboxes1[i, 2], bboxes2[:, 2])
        y_end = np.minimum(bboxes1[i, 3], bboxes2[:, 3])
        overlap = np.maximum(x_end - x_start + 1, 0) * np.maximum(
            y_end - y_start + 1, 0)
        if mode == 'iou':
            union = area1[i] + area2 - overlap
        else:
            union = area1[i] if not exchange else area2
        ious[i, :] = overlap / union
    if exchange:
        ious = ious.T
    return ious

def py_cpu_nms(dets, thresh):
    """Pure Python NMS baseline."""
    # 所有图片的坐标信息，字典形式储存？？
    x1 = dets[:, 0]
    y1 = dets[:, 1]
    x2 = dets[:, 2]
    y2 = dets[:, 3]
    scores = dets[:, 4]

    areas = (x2 - x1 + 1) * (y2 - y1 + 1)  # 计算出所有图片的面积
    order = scores.argsort()[::-1]  # 图片评分按升序排序

    keep = []  # 用来存放最后保留的图片的相应评分
    while order.size > 0:
        i = order[0]  # i 是还未处理的图片中的最大评分
        keep.append(i)  # 保留改图片的值
        # 矩阵操作，下面计算的是图片i分别与其余图片相交的矩形的坐标
        tmp=x1[order[1:]]
        xxxx = x1[i]
        xx1 = np.maximum(x1[i], x1[order[1:]])
        yy1 = np.maximum(y1[i], y1[order[1:]])
        xx2 = np.minimum(x2[i], x2[order[1:]])
        yy2 = np.minimum(y2[i], y2[order[1:]])

        # 计算出各个相交矩形的面积
        w = np.maximum(0.0, xx2 - xx1 + 1)
        h = np.maximum(0.0, yy2 - yy1 + 1)
        inter = w * h
        # 计算重叠比例
        ovr = inter / (areas[i] + areas[order[1:]] - inter)

        # 只保留比例小于阙值的图片，然后继续处理
        inds = np.where(ovr <= thresh)[0]
        indsd= inds+1
        order = order[inds + 1]

    return keep

def soft_nms(boxes, sigma=0.5, Nt=0.1, threshold=0.001, method=1):
    N = boxes.shape[0]
    pos = 0
    maxscore = 0
    maxpos = 0

    for i in range(N):
        maxscore = boxes[i, 4]
        maxpos = i

        tx1 = boxes[i,0]
        ty1 = boxes[i,1]
        tx2 = boxes[i,2]
        ty2 = boxes[i,3]
        ts = boxes[i,4]

        pos = i + 1
    # get max box
        while pos < N:
            if maxscore < boxes[pos, 4]:
                maxscore = boxes[pos, 4]
                maxpos = pos
            pos = pos + 1

    # add max box as a detection
        boxes[i,0] = boxes[maxpos,0]
        boxes[i,1] = boxes[maxpos,1]
        boxes[i,2] = boxes[maxpos,2]
        boxes[i,3] = boxes[maxpos,3]
        boxes[i,4] = boxes[maxpos,4]

    # swap ith box with position of max box
        boxes[maxpos,0] = tx1
        boxes[maxpos,1] = ty1
        boxes[maxpos,2] = tx2
        boxes[maxpos,3] = ty2
        boxes[maxpos,4] = ts

        tx1 = boxes[i,0]
        ty1 = boxes[i,1]
        tx2 = boxes[i,2]
        ty2 = boxes[i,3]
        ts = boxes[i,4]

        pos = i + 1
    # NMS iterations, note that N changes if detection boxes fall below threshold
        while pos < N:
            x1 = boxes[pos, 0]
            y1 = boxes[pos, 1]
            x2 = boxes[pos, 2]
            y2 = boxes[pos, 3]
            s = boxes[pos, 4]

            area = (x2 - x1 + 1) * (y2 - y1 + 1)
            iw = (min(tx2, x2) - max(tx1, x1) + 1)
            if iw > 0:
                ih = (min(ty2, y2) - max(ty1, y1) + 1)
                if ih > 0:
                    ua = float((tx2 - tx1 + 1) * (ty2 - ty1 + 1) + area - iw * ih)
                    ov = iw * ih / ua #iou between max box and detection box

                    if method == 1: # linear
                        if ov > Nt:
                            weight = 1 - ov
                        else:
                            weight = 1
                    elif method == 2: # gaussian
                        weight = np.exp(-(ov * ov)/sigma)
                    else: # original NMS
                        if ov > Nt:
                            weight = 0
                        else:
                            weight = 1

                    boxes[pos, 4] = weight*boxes[pos, 4]
                    print(boxes[:, 4])

            # if box score falls below threshold, discard the box by swapping with last box
            # update N
                    if boxes[pos, 4] < threshold:
                        boxes[pos,0] = boxes[N-1, 0]
                        boxes[pos,1] = boxes[N-1, 1]
                        boxes[pos,2] = boxes[N-1, 2]
                        boxes[pos,3] = boxes[N-1, 3]
                        boxes[pos,4] = boxes[N-1, 4]
                        N = N - 1
                        pos = pos - 1

            pos = pos + 1
    keep = [i for i in range(N)]
    return keep




# -*- coding: UTF-8 -*-
import cv2
import numpy as np
from ..utils import is_str,is_seq,check_file_exist,USE_OPENCV2,mkdir,np_to_base64
import os
from PIL import Image

if not USE_OPENCV2:
    from cv2 import IMREAD_COLOR, IMREAD_GRAYSCALE, IMREAD_UNCHANGED
else:
    from cv2 import CV_LOAD_IMAGE_COLOR as IMREAD_COLOR
    from cv2 import CV_LOAD_IMAGE_GRAYSCALE as IMREAD_GRAYSCALE
    from cv2 import CV_LOAD_IMAGE_UNCHANGED as IMREAD_UNCHANGED

imread_flags = {
    'color': IMREAD_COLOR,
    'grayscale': IMREAD_GRAYSCALE,
    'unchanged': IMREAD_UNCHANGED
}

def imlist(img_dir, valid_exts=None, if_recursive=False):
    """
    List images under directory
    :param img_dir:
    :param valid_exts:
    :param if_recursive:
    :return:
    """
    from glob import glob
    if is_str(valid_exts):
        valid_exts = [valid_exts.strip(".")]
    valid_exts = list(valid_exts) if is_seq(valid_exts) else ["jpg","jpeg","bmp","tif","gif","png"]
    images = []
    for ext in valid_exts:
        images.extend(glob(os.path.join(img_dir,"**","*.%s" % ext),recursive=if_recursive))
    return images

def imread(img_or_path, flag='color'):
    """Read an image.
    Args:
        img_or_path (ndarray or str or pillow Image): Either a numpy array or image path.
            If it is a numpy array (loaded image), then it will be returned
            as is.
        flag (str): Flags specifying the color type of a loaded image,
            candidates are `color`, `grayscale` and `unchanged`.
    Returns:
        ndarray: Loaded image array.
    """
    if isinstance(img_or_path, np.ndarray):
        return img_or_path
    elif is_str(img_or_path):
        flag = imread_flags[flag] if is_str(flag) else flag
        check_file_exist(img_or_path,
                         'img file does not exist: {}'.format(img_or_path))
        image = cv2.imread(img_or_path, flag)
        if flag != 0:
            image = image[...,::-1]
        return image
    else:
        try:
            return pil_img_to_np(img_or_path)
        except:
            raise TypeError('"img" must be a numpy array or a filename')

def imread_topil(img_or_path):
    if is_str(img_or_path):
        return Image.open(img_or_path)
    elif isinstance(img_or_path, np.ndarray):
        return np_img_to_pil(img_or_path)
    elif Image.isImageType(img_or_path):
        return img_or_path
    else:
        raise TypeError('"img" must be a numpy array or a filename or a PIL Image')

def imread_tob64(img_or_path):
    img = imread(img_or_path)
    img_b64 = np_to_base64(img)
    return img_b64.decode("utf-8")

def imwrite(img, file_path, params=None, auto_mkdir=True):
    """Write image to file
    Args:
        img (ndarray): Image array to be written.
        file_path (str): Image file path.
        params (None or list): Same as opencv's :func:`imwrite` interface.
        auto_mkdir (bool): If the parent folder of `file_path` does not exist,
            whether to create it automatically.
    Returns:
        bool: Successful or not.
    """
    if auto_mkdir:
        dir_name = os.path.abspath(os.path.dirname(file_path))
        mkdir(dir_name)

    img = imread(img)
    return np_img_to_pil(img).save(file_path)

    # return cv2.imwrite(file_path, img, params)

def pil_img_to_np(image):
    '''
    将图片加载为numpy数组
    :param image: PIL图片
    :return:
    '''
    (im_width, im_height) = image.size
    dim = 1 if len(np.array(image).shape) == 2 else np.array(image).shape[-1]
    return np.array(image.getdata()).reshape(
        (im_height, im_width, dim)).astype(np.uint8)


def np_img_to_pil(np_arr):
    '''
    将numpy array转换为图片
    :param np_arr:
    :return:
    '''
    im = Image.fromarray(np_arr.astype('uint8')).convert('RGB')
    return im




# -*- coding: utf-8 -* -
import tensorflow as tf

def reframe_box_masks_to_image_masks(box_masks, boxes, image_height,
                                     image_width):
    """Transforms the box masks back to full image masks.
    Embeds masks in bounding boxes of larger masks whose shapes correspond to
    image shape.
    Args:
      box_masks: A tf.float32 tensor of size [num_masks, mask_height, mask_width].
      boxes: A tf.float32 tensor of size [num_masks, 4] containing the box
             corners. Row i contains [ymin, xmin, ymax, xmax] of the box
             corresponding to mask i. Note that the box corners are in
             normalized coordinates.
      image_height: Image height. The output mask will have the same height as
                    the image height.
      image_width: Image width. The output mask will have the same width as the
                   image width.
    Returns:
      A tf.float32 tensor of size [num_masks, image_height, image_width].
    """
    def reframe_box_masks_to_image_masks_default():
        """The default function when there are more than 0 box masks."""

        def transform_boxes_relative_to_boxes(boxes, reference_boxes):
            boxes = tf.reshape(boxes, [-1, 2, 2])
            min_corner = tf.expand_dims(reference_boxes[:, 0:2], 1)
            max_corner = tf.expand_dims(reference_boxes[:, 2:4], 1)
            transformed_boxes = (boxes - min_corner) / (max_corner - min_corner)
            return tf.reshape(transformed_boxes, [-1, 4])

        box_masks_expanded = tf.expand_dims(box_masks, axis=3)
        num_boxes = tf.shape(box_masks_expanded)[0]
        unit_boxes = tf.concat(
            [tf.zeros([num_boxes, 2]), tf.ones([num_boxes, 2])], axis=1)
        reverse_boxes = transform_boxes_relative_to_boxes(unit_boxes, boxes)
        return tf.image.crop_and_resize(
            image=box_masks_expanded,
            boxes=reverse_boxes,
            box_ind=tf.range(num_boxes),
            crop_size=[image_height, image_width],
            extrapolation_value=0.0)

    image_masks = tf.cond(
        tf.shape(box_masks)[0] > 0,
        reframe_box_masks_to_image_masks_default,
        lambda: tf.zeros([0, image_height, image_width, 1], dtype=tf.float32))
    return tf.squeeze(image_masks, axis=3)



# -*- coding: UTF-8 -*-
from ..labelme import LabelMe
from ..coco import Coco
from ..voc import Voc
from ..core.sample import Sample
from icv.utils import reset_dir, make_empty_coco_anno, mkfile
import json
import os


class LabelMeConverter(object):
    def __init__(self, labelme):
        assert isinstance(labelme, LabelMe)
        self.labelme = labelme

    def to_coco(self, coco_root, split=None,reset=False):
        split = split if split is not None else self.labelme.split
        if reset:
            reset_dir(coco_root)
        dist_anno_path, dist_image_path = Coco.reset_dir(coco_root, split)
        dist_anno_file = os.path.join(dist_anno_path, "%s.json" % split)

        anno_data = make_empty_coco_anno()
        json.dump(anno_data, open(dist_anno_file, "w+"))

        coco = Coco(
            dist_image_path,
            dist_anno_file,
            keep_no_anno_image=self.labelme.keep_no_anno_image,
            one_index=self.labelme.one_index,
        )
        coco.set_categories(self.labelme.categories)

        image_id = 1

        for sample in self.labelme.get_samples():
            assert isinstance(sample, Sample)
            coco.sample_db[image_id] = sample
            coco.ids.append(image_id)
            image_id += 1

        coco.save(coco_root, reset_dir=True, split=split)
        coco.init()
        return coco

    def to_voc(self, voc_root, split=None,reset=False):
        split = split if split is not None else self.labelme.split
        if reset:
            reset_dir(voc_root)
        anno_path, image_path, imgset_path, imgset_seg_path, \
        seg_class_image_path, seg_object_image_path = Voc.reset_dir(voc_root)

        setfile = os.path.join(imgset_seg_path, "%s.txt" % split) if self.labelme.is_seg_mode else os.path.join(
            imgset_path, "%s.txt" % split)
        mkfile(setfile)

        voc = Voc(
            voc_root,
            split,
            keep_no_anno_image=self.labelme.keep_no_anno_image,
            mode="detect" if not self.labelme.is_seg_mode else "segment",
            categories=self.labelme.categories,
            one_index=self.labelme.one_index,
        )

        for sample in self.labelme.get_samples():
            assert isinstance(sample, Sample)
            voc.sample_db[sample.name] = sample
            voc.ids.append(sample.name)

        voc.save(voc_root, reset_dir=True, split=split)
        voc.init()
        return voc



# -*- coding: UTF-8 -*-
import torch
import numpy as np
from ..utils import is_seq, is_file, is_str, labelmap_to_category_index
from abc import ABCMeta, abstractmethod
from .service.server import DetectorServer


class Detector(object):
    __metaclass__ = ABCMeta

    def __init__(self, categories=None, labelmap_path=None, iou_thr=0.5, score_thr=0.5, device=None):
        assert is_seq(categories) or is_file(
            labelmap_path), "param categories and param labelmap_path should input at_least one."
        assert iou_thr > 0 and iou_thr < 1, "param iou_thr should > 0 and < 1."
        assert score_thr > 0 and score_thr < 1, "param score_thr should > 0 and < 1."

        self.categories = categories
        self.labelmap_path = labelmap_path
        self.iou_thr = iou_thr
        self.score_thr = score_thr

        self.device = device if device else torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

        self._labelmap()

    def _labelmap(self):
        if self.categories is not None:
            self.category_index = {id + 1: {"id": id + 1, "name": label} for id, label in enumerate(self.categories)}
            self.category_name_index = {self.category_index[cat_id]["name"]: self.category_index[cat_id] for cat_id in
                                        self.category_index}
        else:
            self.category_index = labelmap_to_category_index(self.labelmap_path,
                                                             use_display_name=True)
            self.category_name_index = {self.category_index[cat_id]["name"]: self.category_index[cat_id] for cat_id in
                                        self.category_index}

    def _warmup(self):
        self.inference(np.zeros((100, 100, 3), dtype=np.uint8))

    @abstractmethod
    def inference(self, image, show=False, save_path=None, score_thr=-1):
        pass

    @abstractmethod
    def inference_batch(self, images, save_dir=None, resize=None, score_thr=-1):
        pass

    @abstractmethod
    def start_server(
            self,
            port=9527,
            open_web=True,
            secret=None,
            upload_dir=None,
            debug=True,
            name="",
            default_params_score_thr=0.5
    ):
        assert isinstance(port, int)
        assert secret is None or is_str(secret)
        server = DetectorServer(
            self,
            secrect=secret,
            open_web=open_web,
            name=name,
            default_params_score_thr=default_params_score_thr
        )
        server.start_server(port=port, upload_dir=upload_dir, debug=debug)



# -*- coding: UTF-8 -*-
from time import time,localtime,strftime

class Time(object):
    def __init__(self,):
        self.running = False

    @staticmethod
    def now():
        return time()

    @staticmethod
    def now_str():
        return strftime("%Y-%m-%d %H:%M:%S", localtime(time()))

    @staticmethod
    def is_before(t):
        return time() - t < 0

    @staticmethod
    def is_after(t):
        return time() - t > 0

    def start(self):
        self.running = True
        self._t_start = time()
        self._t_last = time()

    def since_start(self):
        if not self.running:
            return 0
        self._t_last = time()
        return self._t_last - self._t_start

    def since_last(self):
        if not self.running:
            return 0
        self._t_last = time()
        return self._t_last - self._t_start




